@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE DefaultSignatures #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE GADTs #-}
@@ -6,12 +7,14 @@ module Utilities where
 
 import Test.QuickCheck
 
+import Control.Arrow ((***))
 import Data.Foldable
 import Data.Bifunctor
 import qualified Data.Vector as DV
 import qualified Data.Vector.Generic as DVG
 import qualified Data.Vector.Primitive as DVP
 import qualified Data.Vector.Storable as DVS
+import qualified Data.Vector.Strict  as DVV
 import qualified Data.Vector.Unboxed as DVU
 import qualified Data.Vector.Fusion.Bundle as S
 
@@ -43,6 +46,12 @@ instance (Arbitrary a, DVS.Storable a) => Arbitrary (DVS.Vector a) where
 instance (CoArbitrary a, DVS.Storable a) => CoArbitrary (DVS.Vector a) where
     coarbitrary = coarbitrary . DVS.toList
 
+instance (Arbitrary a) => Arbitrary (DVV.Vector a) where
+    arbitrary = fmap DVV.fromList arbitrary
+
+instance (CoArbitrary a) => CoArbitrary (DVV.Vector a) where
+    coarbitrary = coarbitrary . DVV.toList
+
 instance (Arbitrary a, DVU.Unbox a) => Arbitrary (DVU.Vector a) where
     arbitrary = fmap DVU.fromList arbitrary
 
@@ -95,6 +104,11 @@ instance (Eq a, DVS.Storable a, TestData a) => TestData (DVS.Vector a) where
   model   = map model    . DVS.toList
   unmodel = DVS.fromList . map unmodel
 
+instance (Eq a, TestData a) => TestData (DVV.Vector a) where
+  type Model (DVV.Vector a) = [Model a]
+  model   = map model    . DVV.toList
+  unmodel = DVV.fromList . map unmodel
+
 instance (Eq a, DVU.Unbox a, TestData a) => TestData (DVU.Vector a) where
   type Model (DVU.Vector a) = [Model a]
   model   = map model    . DVU.toList
@@ -292,6 +306,12 @@ imapMaybe f = catMaybes . withIndexFirst map f
 
 indexedLeftFold fld f z = fld (uncurry . f) z . zip [0..]
 
+spanR :: (a -> Bool) -> [a] -> ([a], [a])
+spanR f = (reverse *** reverse) . span f . reverse
+
+breakR :: (a -> Bool) -> [a] -> ([a], [a])
+breakR f = (reverse *** reverse) . break f . reverse
+
 ifoldl :: (a -> Int -> a -> a) -> a -> [a] -> a
 ifoldl = indexedLeftFold foldl
 
