@@ -25,6 +25,9 @@ main = mapM_ run modGroups
       , [ "src/Data/Vector.hs"
         , "src/Data/Vector/Mutable.hs"
         ]
+      , [ "src/Data/Vector/Strict.hs"
+        , "src/Data/Vector/Strict/Mutable.hs"
+        ]
       , [ "src/Data/Vector/Generic.hs"
         , "src/Data/Vector/Generic/Mutable.hs"
         ]
